;;; EXCEPTION.SCM
;;;
;;; A simple exception Scheme system. You DO NOT have to understand this code
;;; at this point of the semester. 


;; ------------------------------------------------------------------------
;; CATCH & THROW

(define catch
  (lambda (try-thunk user-handler)
    (call-with-current-continuation 
     (lambda (catch-cont)
       (begin
         ;; Push handler on stack
         (push-handler! (make-catch-handler user-handler catch-cont))
         ;; Evaluate try-thunk
         (let ((try-value (try-thunk)))
           ;; In normal case, pop handler
           (begin
             (pop-handler!) ;; Remove handler from stack
             try-value))))))) ;; Return value

(define throw
  ;; TAG is a symbol indicating the type of exception.
  ;; VALUE is the "argument" of the exception.
  (lambda (tag value)
    (if (null? *handler-stack*)
      ; Treat unhandled THROW like Scheme's ERROR
      (error (string-append "Unhandled exception " 
                            (symbol->string tag)
                            ": ") 
             value) 
      ((pop-handler!) tag value))))

;; ------------------------------------------------------------------------
;; The global handler stack 

(define *handler-stack* '())

(define push-handler!
  (lambda (catch-handler)
    (set! *handler-stack* (cons catch-handler *handler-stack*))))

(define pop-handler! 
  (lambda () 
    (if (null? *handler-stack*)
       (error "POP-HANDLER!: empty stack.")
       (let ((top (car *handler-stack*)))
         (begin
           (set! *handler-stack* (cdr *handler-stack*))
           top)))))

(define make-catch-handler 
  (lambda (user-handler catch-cont)
    (lambda (tag value)
       ;; Invoking CATCH-CONT returns directly to nearest enclosing CATCH,
       ;; ignoring the current continuation.
       (catch-cont (user-handler tag value)))))

;; ------------------------------------------------------------------------
;; Test cases

(define catch-test1
  (lambda (x)
    (catch (lambda () 
             (* (+ (if (even? x) 
                     (throw 'even x)
                     x)
                   1)
                (- (if (< x 0) 
                     (throw 'negative x)
                     x)
                   1)))
           (lambda (tag value) 
             (display (string-append "Caught exception " 
                                     (symbol->string tag)))
             (newline)
             value))))

