;;; NAMESPACE.SCM
;;; Use MIT Scheme's environments to allow 
;;; different programs with the same names 
;;; (e.g. toy language interpreters) to co-exist. 

(if (not (environment-bound? user-initial-environment '*namespaces*))
    (begin 
      (eval '(define *namespaces* (list (list 'user user-initial-environment)))
	    user-initial-environment)
      (eval '(define *current-namespace* 'user)
	    user-initial-environment)))

(define (namespace) *current-namespace*)

(define (namespaces) (map first *namespaces*))

(define (namespace? name) 
  (not (null? (assq name *namespaces*))))

(define (name-to-namespace name)
  (let ((probe (assq name *namespaces*)))
    (if (not (null? probe))
	(second probe)
	(error "Not a legal namespace name: " name))))

(define (goto-namespace name)
  (if (eq? name (namespace))
      (display (string-append "Already in namespace " 
			      (symbol->string name)
			      "\n"))
      (let ((ns (name-to-namespace name)))
	(display (string-append "Leaving namespace "
				(symbol->string (namespace))
				"\n"))
	(display (string-append "Entering namespace "
				(symbol->string name)
				"\n"))
	(set! *current-namespace* name)
	(ge ns))))

(define (user) (goto-namespace 'user))

(define (new-namespace name)
  ;; If namespace with name already exists, do nothing
  ;; Otherwise, create a new namespace with the given name,
  ;; and create a nullary function with the same name in 
  ;; user-initial-environment that goes to that namespace. 
  (if (not (namespace? name))
      (let ((ns (eval '(make-environment) user-initial-environment)))
	(begin 
	  (set! *namespaces* (cons (list name ns) *namespaces*))
	  (eval `(define ,name (lambda () (goto-namespace ',name)))
		user-initial-environment)))))

(define (export sym value)
  ;; export value from current namespace to user-initial-environment 
  ;; with name SYM
  (eval `(define ,sym ,value) user-initial-environment)
  )
	





