;;; SET-LIST-NODUPS.SCM
;;; Implementation of set ADT in terms of lists without duplicates

; (load "../Scheme-utilities/list-ops.scm")

(define set-empty
  (lambda () '()))

(define set-empty?
  (lambda (s) (null? s)))

(define set-singleton
  (lambda (x) (list x)))

(define set-member? 
  (lambda (x s)
    (member x s)))

(define set-insert
  (lambda (x s)
    (if (set-member? x s)
      s
      (cons x s))))

(define set-choose
  (lambda (s) 
    (if (null? s)
      (ERROR "Can't choose element from empty set")
      (car s))))

(define set-union
  (lambda (s1 s2)
    (if (null? s1)
      s2
      (set-insert (car s1) 
                  (set-union (cdr s1) s2)))))

(define set-intersection
  (lambda (s1 s2)
    (if (null? s1)
      '()
      (if (member (car s1) s2)
        (cons (car s1) (set-intersection (cdr s1) s2))
        (set-intersection (cdr s1) s2)))))

(define set-difference
  (lambda (s1 s2)
    (if (null? s1)
      '()
      (if (member (car s1) s2)
        (set-difference (cdr s1) s2)
        (cons (car s1) (set-difference (cdr s1) s2))))))

(define set-subset?
  (lambda (s1 s2)
    (forall? (lambda (x1) (set-member? x1 s2)) s1)))

(define set-equal?
  (lambda (s1 s2)
    (and (set-subset? s1 s2)
	 (set-subset? s2 s1))))

(define list->set
  (lambda (lst)
    (if (null? lst)
      '()
      (set-insert (car lst)
                  (list->set (cdr lst))))))

(define set->list
  (lambda (set) set))

(define set-map
  (lambda (f s)
    (map f s)))

(define set-filter
  (lambda (p s)
    (filter p s)))

(define set-fold
  ;; Should use a binop that is order indepedent,
  ;; since order of elements in fold cannot be predicted.
  (lambda (binop null s)
    (foldr binop null s)))

(define set-union-map 
  (lambda (f xs)
    (foldr set-union (set-empty) (map f xs))))
